from asyncio import CancelledError

ALLOWED_EXCEPTIONS = (
    CancelledError,
    KeyboardInterrupt,
    SystemExit,
    GeneratorExit,
)


def require(result, message: str = None) -> None:
    if not result:
        raise ValueError(message) if message else ValueError()



from argparse import Namespace
from typing import Literal

from rosy import Node
from rosy.argparse import get_node_arg_parser
from rosy.authentication import Authenticator, optional_authkey_authenticator
from rosy.codec import (
    Codec,
    DictCodec,
    FixedLengthIntCodec,
    LengthPrefixedStringCodec,
    SequenceCodec,
    json_codec,
    msgpack_codec,
    pickle_codec,
)
from rosy.coordinator.client import build_coordinator_client
from rosy.coordinator.constants import DEFAULT_COORDINATOR_PORT
from rosy.network import get_lan_hostname
from rosy.node.clienthandler import ClientHandler
from rosy.node.codec import NodeMessageCodec
from rosy.node.loadbalancing import (
    GroupingTopicLoadBalancer,
    LeastRecentLoadBalancer,
    ServiceLoadBalancer,
    TopicLoadBalancer,
    node_name_group_key,
)
from rosy.node.peer import PeerConnectionBuilder, PeerConnectionManager, PeerSelector
from rosy.node.servers import ServerProvider, ServersManager, TcpServerProvider, TmpUnixServerProvider
from rosy.node.service.caller import ServiceCaller
from rosy.node.service.codec import ServiceRequestCodec, ServiceResponseCodec
from rosy.node.service.handlermanager import ServiceHandlerManager
from rosy.node.service.requesthandler import ServiceRequestHandler
from rosy.node.topic.codec import TopicMessageCodec
from rosy.node.topic.listenermanager import TopicListenerManager
from rosy.node.topic.messagehandler import TopicMessageHandler
from rosy.node.topic.sender import TopicSender
from rosy.node.topology import MeshTopologyManager
from rosy.specs import NodeId
from rosy.types import Data, Host, Port, ServerHost

DataCodecArg = Codec[Data] | Literal['pickle', 'json', 'msgpack']


async def build_node_from_args(
        default_node_name: str = None,
        args: Namespace = None,
        **kwargs,
) -> Node:
    """
    Builds a node from command line arguments.

    Args:
        default_node_name:
            Default node name. If not given, the argument is required.
            Ignored if `args` is given.
        args:
            Arguments from an argument parser. If not given, an argument parser
            is created using `get_node_arg_parser` and is used to parse args.
            This is useful if you create your own argument parser.
        kwargs:
            Additional keyword arguments to pass to `build_node`.
            These will override anything specified in `args`.
    """

    if args is None:
        args = get_node_arg_parser(default_node_name).parse_args()

    build_args = vars(args)

    if hasattr(args, 'coordinator'):
        build_args['coordinator_host'] = args.coordinator.host
        build_args['coordinator_port'] = args.coordinator.port

    build_args.update(kwargs)

    return await build_node(**build_args)


async def build_node(
        name: str,
        coordinator_host: Host = 'localhost',
        coordinator_port: Port = DEFAULT_COORDINATOR_PORT,
        coordinator_reconnect_timeout: float | None = 5.0,
        allow_unix_connections: bool = True,
        allow_tcp_connections: bool = True,
        node_server_host: ServerHost = None,
        node_client_host: Host = None,
        data_codec: DataCodecArg = 'pickle',
        authkey: bytes = None,
        authenticator: Authenticator = None,
        topic_load_balancer: TopicLoadBalancer = None,
        service_load_balancer: ServiceLoadBalancer = None,
        start: bool = True,
        **kwargs,
) -> Node:
    """Builds a ROSY node.

    Args:
        name: Name of the node. Does not need to be unique; multiple nodes
            with the same name can be part of the same mesh. Topic messages
            will be sent to nodes of the same name in a round-robin fashion,
            by default, according to the value of `topic_load_balancer`.
            This makes horizontal scaling easy; just start the same node
            multiple times.
        coordinator_host: Hostname of the coordinator. Defaults to 'localhost'.
        coordinator_port: Port of the coordinator. Defaults to 7679.
        coordinator_reconnect_timeout: Time in seconds to wait before
            reconnecting to the coordinator. Defaults to 5.0.
        allow_unix_connections: Whether to allow connections to the node over
            Unix sockets. Defaults to True.
        allow_tcp_connections: Whether to allow connections to the node over
            TCP sockets. Defaults to True.
        node_server_host: Hostname to use for the node's server. If not given,
            the node will listen on all available network interfaces.
        node_client_host: Hostname that other nodes will use to connect to this
            node. If not given, defaults to the machine's mDNS hostname, e.g.
            "<hostname>.local".
        data_codec: A codec to use for serializing and deserializing data
            between nodes. Can be one of 'pickle', 'json', or 'msgpack';
            or, a custom Codec instance. Defaults to 'pickle'.
        authkey: Authentication key that nodes will use to authenticate each
            other. Defaults to None (no authentication).
        authenticator: An authenticator to use for the node. If not given, the
            node will use an authenticator according to the value of `authkey`.
        topic_load_balancer: A load balancer to use for distributing topic
            messages. Defaults to a least-recently-used load balancer.
        service_load_balancer: A load balancer to use for distributing service
            requests. Defaults to a least-recently-used load balancer.
        start: Whether to start the node immediately. Defaults to True.
            If False, the user must call `await node.start()` before the node
            will be ready to use.
    """

    authenticator = authenticator or optional_authkey_authenticator(authkey)

    coordinator_client = await build_coordinator_client(
        coordinator_host,
        coordinator_port,
        authenticator,
        reconnect_timeout=coordinator_reconnect_timeout,
    )

    topic_listener_manager = TopicListenerManager()
    service_handler_manager = ServiceHandlerManager()

    request_id_bytes = 2
    node_message_codec = build_node_message_codec(request_id_bytes, data_codec)

    servers_manager = build_servers_manager(
        allow_unix_connections,
        allow_tcp_connections,
        node_server_host,
        node_client_host,
        authenticator,
        topic_listener_manager,
        service_handler_manager,
        node_message_codec,
    )

    topology_manager = MeshTopologyManager()

    peer_selector = build_peer_selector(
        topology_manager,
        topic_load_balancer,
        service_load_balancer,
    )

    connection_manager = PeerConnectionManager(
        PeerConnectionBuilder(authenticator),
    )

    topic_sender = TopicSender(peer_selector, connection_manager, node_message_codec)

    service_caller = ServiceCaller(
        peer_selector,
        connection_manager,
        node_message_codec,
        max_request_ids=2 ** (8 * request_id_bytes),
    )

    node = Node(
        id=NodeId(name),
        coordinator_client=coordinator_client,
        servers_manager=servers_manager,
        topology_manager=topology_manager,
        connection_manager=connection_manager,
        topic_sender=topic_sender,
        topic_listener_manager=topic_listener_manager,
        service_caller=service_caller,
        service_handler_manager=service_handler_manager,
    )

    if start:
        await node.start()

    return node


def build_node_message_codec(
        request_id_bytes: int,
        data_codec: DataCodecArg,
) -> NodeMessageCodec:
    data_codec = build_data_codec(data_codec)

    short_string_codec = LengthPrefixedStringCodec(
        len_prefix_codec=FixedLengthIntCodec(length=1)
    )

    short_int_codec = FixedLengthIntCodec(length=1)

    args_codec: SequenceCodec[Data] = SequenceCodec(
        len_header_codec=short_int_codec,
        item_codec=data_codec,
    )

    kwargs_codec: DictCodec[str, Data] = DictCodec(
        len_header_codec=short_int_codec,
        key_codec=short_string_codec,
        value_codec=data_codec,
    )

    request_id_codec = FixedLengthIntCodec(length=request_id_bytes)

    return NodeMessageCodec(
        topic_message_codec=TopicMessageCodec(
            topic_codec=short_string_codec,
            args_codec=args_codec,
            kwargs_codec=kwargs_codec,
        ),
        service_request_codec=ServiceRequestCodec(
            request_id_codec,
            service_codec=short_string_codec,
            args_codec=args_codec,
            kwargs_codec=kwargs_codec,
        ),
        service_response_codec=ServiceResponseCodec(
            request_id_codec,
            data_codec=data_codec,
            error_codec=LengthPrefixedStringCodec(
                len_prefix_codec=FixedLengthIntCodec(length=2),
            )
        ),
    )


def build_data_codec(data_codec: DataCodecArg) -> Codec[Data]:
    if data_codec == 'pickle':
        return pickle_codec
    elif data_codec == 'json':
        return json_codec
    elif data_codec == 'msgpack':
        return msgpack_codec
    else:
        return data_codec


def build_servers_manager(
        allow_unix_connections: bool,
        allow_tcp_connections: bool,
        node_server_host: ServerHost,
        node_client_host: Host | None,
        authenticator: Authenticator,
        topic_listener_manager: TopicListenerManager,
        service_handler_manager: ServiceHandlerManager,
        node_message_codec: NodeMessageCodec,
) -> ServersManager:
    topic_message_handler = TopicMessageHandler(topic_listener_manager)

    service_request_handler = ServiceRequestHandler(
        service_handler_manager,
        node_message_codec,
    )

    client_handler = ClientHandler(
        authenticator,
        node_message_codec,
        topic_message_handler,
        service_request_handler,
    )

    server_providers = build_server_providers(
        allow_unix_connections,
        allow_tcp_connections,
        node_server_host,
        node_client_host,
    )

    return ServersManager(server_providers, client_handler.handle_client)


def build_server_providers(
        allow_unix_connections: bool,
        allow_tcp_connections: bool,
        node_server_host: ServerHost | None,
        node_client_host: Host | None,
) -> list[ServerProvider]:
    server_providers = []

    if allow_unix_connections:
        server_providers.append(TmpUnixServerProvider())

    if allow_tcp_connections:
        if not node_client_host:
            node_client_host = get_lan_hostname()

        provider = TcpServerProvider(node_server_host, node_client_host)
        server_providers.append(provider)

    if not server_providers:
        raise ValueError('Must allow at least one type of connection')

    return server_providers


def build_peer_selector(
        topology_manager: MeshTopologyManager,
        topic_load_balancer: TopicLoadBalancer | None,
        service_load_balancer: ServiceLoadBalancer | None,
) -> PeerSelector:
    least_recent_load_balancer = LeastRecentLoadBalancer()

    default_topic_load_balancer = GroupingTopicLoadBalancer(
        group_key=node_name_group_key,
        load_balancer=least_recent_load_balancer,
    )

    return PeerSelector(
        topology_manager,
        topic_load_balancer=topic_load_balancer or default_topic_load_balancer,
        service_load_balancer=service_load_balancer or least_recent_load_balancer,
    )



from argparse import ArgumentParser, Namespace
from typing import Iterable

from rosy.authentication import optional_authkey_authenticator
from rosy.coordinator.client import build_coordinator_client
from rosy.specs import MeshTopologySpec


def add_log_arg(parser: ArgumentParser, default: str = 'WARNING') -> None:
    parser.add_argument(
        '--log',
        default=default,
        help='Log level; DEBUG, INFO, ERROR, etc. Default: %(default)s'
    )


async def get_mesh_topology(args: Namespace) -> MeshTopologySpec:
    authenticator = optional_authkey_authenticator(args.authkey)

    coordinator_client = await build_coordinator_client(
        host=args.coordinator.host,
        port=args.coordinator.port,
        authenticator=authenticator,
        reconnect_timeout=None,
    )

    return await coordinator_client.get_topology()


def print_args_and_kwargs(args: Iterable, kwargs: dict) -> None:
    if args:
        print('args:')
        for i, arg in enumerate(args):
            arg = arg_to_str(arg)
            print(f'  {i}: {arg}')

    if kwargs:
        print('kwargs:')
        for key, value in kwargs.items():
            value = arg_to_str(value)
            print(f'  {key}={value}')


def arg_to_str(arg) -> str:
    arg = repr(arg)

    if '\n' in arg:
        arg = f'```\n{arg}\n```'

    return arg



import asyncio
import logging
from abc import ABC, abstractmethod
from asyncio import StreamReader
from codecs import StreamWriter

from rosy.asyncio import close_ignoring_errors
from rosy.authentication import AuthKey, Authenticator, optional_authkey_authenticator
from rosy.coordinator.constants import DEFAULT_COORDINATOR_HOST, DEFAULT_COORDINATOR_PORT
from rosy.objectio import CodecObjectReader, CodecObjectWriter, ObjectIO
from rosy.reqres import MeshTopologyBroadcast, RegisterNodeRequest, RegisterNodeResponse
from rosy.rpc import ObjectIORPC, RPC
from rosy.specs import MeshNodeSpec, MeshTopologySpec, NodeId
from rosy.types import Port, ServerHost
from rosy.utils import ALLOWED_EXCEPTIONS

logger = logging.getLogger(__name__)


class MeshCoordinatorServer(ABC):
    @abstractmethod
    async def start(self) -> None:
        ...


class RPCMeshCoordinatorServer(MeshCoordinatorServer):
    def __init__(
            self,
            start_stream_server,
            build_rpc,
            authenticator: Authenticator,
    ):
        self.start_stream_server = start_stream_server
        self.build_rpc = build_rpc
        self.authenticator = authenticator

        self._node_clients: dict[RPC, NodeId | None] = {}
        self._nodes: dict[NodeId, MeshNodeSpec] = {}

    async def start(self) -> None:
        server = await self.start_stream_server(self._handle_connection)

    async def _handle_connection(self, reader: StreamReader, writer: StreamWriter) -> None:
        peer_name = writer.get_extra_info('peername') or writer.get_extra_info('sockname')
        logger.debug(f'New connection from: {peer_name}')

        await self.authenticator.authenticate(reader, writer)

        rpc = self.build_rpc(reader, writer)
        rpc.request_handler = lambda r: self._handle_request(r, rpc, peer_name)
        self._node_clients[rpc] = None

        try:
            await rpc.run_forever()
        except (ConnectionResetError, EOFError):
            logger.debug(f'Client disconnected: {peer_name}')
        finally:
            try:
                await self._remove_node(rpc)
            finally:
                await close_ignoring_errors(writer)

    async def _handle_request(self, request, rpc: RPC, peer_name: str):
        if request == b'get_topology':
            return self._get_mesh_topology()
        elif request == b'ping':
            logger.debug(f'Received heartbeat from: {peer_name}')
            return b'pong'
        elif isinstance(request, RegisterNodeRequest):
            return await self._handle_register_node(request, rpc)
        else:
            raise Exception(f'Received invalid request object of type={type(request)}')

    def _get_mesh_topology(self) -> MeshTopologySpec:
        nodes = sorted(self._nodes.values(), key=lambda n: n.id)
        return MeshTopologySpec(nodes=nodes)

    async def _handle_register_node(
            self,
            request: RegisterNodeRequest,
            rpc: RPC,
    ) -> RegisterNodeResponse:
        logger.debug(f'Got register node request: {request}')

        node_spec = request.node_spec
        node_is_new = node_spec.id not in self._nodes
        self._node_clients[rpc] = node_spec.id
        self._nodes[node_spec.id] = node_spec

        if node_is_new:
            logger.info(f'Node registered: {node_spec.id}')
            logger.info(f'Total nodes: {len(self._nodes)}')
        else:
            logger.debug(f'Node re-registered: {node_spec.id}')

        await self._broadcast_topology()

        return RegisterNodeResponse()

    async def _remove_node(self, rpc: RPC) -> None:
        node_id = self._node_clients.pop(rpc)
        if node_id is None:
            return

        self._nodes.pop(node_id, None)
        logger.info(f'Node removed: {node_id}')
        logger.info(f'Total nodes: {len(self._nodes)}')

        await self._broadcast_topology()

    async def _broadcast_topology(self) -> None:
        logger.debug(f'Broadcasting topology to {len(self._nodes)} nodes...')
        if not self._nodes:
            return

        mesh_topology = self._get_mesh_topology()
        for node in mesh_topology.nodes:
            logger.debug(f'- {node.id}: {node}')

        message = MeshTopologyBroadcast(mesh_topology)

        [
            await self._send_message(node_client, message)
            for node_client in self._node_clients.keys()
        ]

    async def _send_message(self, client: RPC, message: MeshTopologyBroadcast) -> None:
        try:
            await client.send_message(message)
        except ALLOWED_EXCEPTIONS:
            raise
        except Exception as e:
            logger.error(f'Error sending message to client: {e!r}')


def build_mesh_coordinator_server(
        host: ServerHost = DEFAULT_COORDINATOR_HOST,
        port: Port = DEFAULT_COORDINATOR_PORT,
        authkey: AuthKey = None,
        authenticator: Authenticator = None,
) -> MeshCoordinatorServer:
    async def start_stream_server(cb):
        return await asyncio.start_server(cb, host=host, port=port)

    def build_rpc(reader, writer):
        return ObjectIORPC(ObjectIO(
            reader=CodecObjectReader(reader),
            writer=CodecObjectWriter(writer),
        ))

    authenticator = authenticator or optional_authkey_authenticator(authkey)

    return RPCMeshCoordinatorServer(
        start_stream_server,
        build_rpc,
        authenticator,
    )



import logging

from rosy.node.codec import NodeMessageCodec
from rosy.node.peer import PeerConnectionManager, PeerSelector
from rosy.node.topic.types import TopicMessage
from rosy.node.types import Args, KWArgs
from rosy.specs import MeshNodeSpec
from rosy.types import Buffer, Topic
from rosy.utils import ALLOWED_EXCEPTIONS

logger = logging.getLogger(__name__)


class TopicSender:
    def __init__(
            self,
            peer_selector: PeerSelector,
            connection_manager: PeerConnectionManager,
            node_message_codec: NodeMessageCodec,
    ):
        self.peer_selector = peer_selector
        self.connection_manager = connection_manager
        self.node_message_codec = node_message_codec

    async def send(self, topic: Topic, args: Args, kwargs: KWArgs) -> None:
        # TODO handle case of self-sending more efficiently

        nodes = self.peer_selector.get_nodes_for_topic(topic)
        if not nodes:
            return

        message = TopicMessage(topic, args, kwargs)
        data = await self.node_message_codec.encode_topic_message(message)

        [
            await self._send_to_one(n, data)
            for n in nodes
        ]

    async def _send_to_one(self, node: MeshNodeSpec, data: Buffer) -> None:
        try:
            connection = await self.connection_manager.get_connection(node)

            async with connection.writer as writer:
                writer.write(data)
                await writer.drain()
        except ALLOWED_EXCEPTIONS:
            raise
        except Exception as e:
            logger.error(
                f'Error sending topic message to node={node.id}: {e!r}',
            )
